#ifndef _BSTACK_H_
#define _BSTACK_H_

/**
 * Implementation of a bounded stack, which overwrites the bottom value
 * when it gets full.
 *
 * Important: pointers to data pushed on the stack must be malloc'ed;
 * no addresses to variables should be used. This is because all data
 * pointed to by stack nodes will be freed when the stack is destroyed.
*/

typedef struct BSNode
{
  void * data;
} BSNode;

typedef struct BoundedStack
{
  unsigned int top;
  BSNode * nodes;
  unsigned int size;
  unsigned int used;
} BoundedStack;

BoundedStack * BSCreate(unsigned int size);
void BSPush(BoundedStack * stack, void * data);
void * BSPop(BoundedStack * stack);
void BSDestroy(BoundedStack * stack);

#endif /* _BSTACK_H_ */
