#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sock.h"
#include "http.h"

char buffer[BUFSIZ]; /* used for sending requests and receiving responses */
const int httplen = 7; /* strlen("http://"); */
int verbose = 0;

/* private function */

int httpSendRequest(int sockfd, const char * type, char * url)
{
  char * pDomain = NULL;
  char * pPath = NULL;
  char * domain = NULL;
  char * newurl = NULL;
  int bytessent = 0;

  if (strncmp("http://", url, httplen) != 0)
    return -1;

  pDomain = url + httplen; /* skip http:// part */
  pPath = strchr(pDomain, '/');

  if (pPath == NULL) /* domain only and no path => append '/' */
  {
    newurl = (char *) malloc(strlen(url) + 2); /* 1 or 2? */
    strcpy(url, newurl);
    strcat(newurl, "/");
    pDomain = newurl + httplen;
    pPath = strchr(pDomain, '/'); /* may be made more efficient later */
  }

  *pPath = '\0'; /* null-terminate domain part for now */

  domain = (char *) malloc(strlen(pDomain) + 1);
  strcpy(domain, pDomain);

  *pPath = '/'; /* restore it as it was */

  memset(buffer, 0, BUFSIZ);
  sprintf(buffer, "%s %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n",
          type, pPath, domain);
  bytessent = sockWriteStr(sockfd, buffer);

  if (newurl != NULL)
    free(newurl);
  free(domain);

  return bytessent;
}

int httpIsHttp(char * url)
{
  if (strncmp("http://", url, httplen) != 0)
    return 0;
  else
    return 1;
}

int httpSendGetRequest(int sockfd, char * url)
{
  return httpSendRequest(sockfd, "GET", url);
}

int httpRecvResponse(int sockfd, char * filename)
{
  int bytesread = 0;
  char * pPayload = NULL;
  int headerlen = 0;
  int payloadlen = 0;
  int code = 0;
  int totaldownloaded = 0;
  float percentdownloaded = 0;

  /* read first bufferful of data, which contains header */

  memset(buffer, 0, BUFSIZ);
  bytesread = sockRead(sockfd, buffer, BUFSIZ - 1);
  totaldownloaded = bytesread;

  /* retrieve the status code of the response */

  pPayload = strchr(buffer, ' ') + 1;
  pPayload = strchr(buffer, ' ') + 1;
  code = atoi(pPayload);

  if (code != 200)
    return code; /* otherwise it segfaults */

  /* get payload size from header */
  pPayload = strstr(buffer, "Content-Length: ");
  pPayload += 16; /* strlen("Content-Length: ") */
  payloadlen = atoi(pPayload);

  /* find header/payload separator (strstr() requires null-termination) */

  buffer[bytesread + 1] = '\0';
  pPayload = strstr(buffer, "\r\n\r\n");
  pPayload += 4; /* skip \r\n\r\n */
  headerlen = pPayload - buffer;
  totaldownloaded -= headerlen;

  /* get what's left of the payload */

  if (bytesread < BUFSIZ) /* we have all the data... payload and header */
  {
    if (bytesread != headerlen) /* payload exists */
    {
      FILE * payloadfile = fopen(filename, "wb");
      fwrite(pPayload, sizeof(char), bytesread - headerlen, payloadfile);
      while ((bytesread = sockRead(sockfd, buffer, BUFSIZ - 1)) > 0)
      {
        if (verbose == 1)
        {
          totaldownloaded += bytesread;
          percentdownloaded = 100 *
                  ((float) totaldownloaded / (float) payloadlen);
          printf("\r%.0f%% downloaded (%d of %d bytes)", percentdownloaded,
                  totaldownloaded, payloadlen);
        }
        fwrite(buffer, sizeof(char), bytesread, payloadfile);
      }
      fflush(payloadfile);
      fclose(payloadfile);
      printf("\n");
    }
  }

  return code;
}

void httpVerboseOn()
{
  verbose = 1;
}

void httpVerboseOff()
{
  verbose = 0;
}
