#include <SDL/SDL.h>
#include <stdlib.h>
#include <string.h>
#include "bstack.h"
#include "history.h"
#include "timer.h"

void rotateLeft(SDL_Surface * image, BoundedStack * history)
{
  const int width = image->w; //image width (original)
  const int height = image->h; //image height (original)
  const int bpp = image->format->BitsPerPixel; //bpp (normally 32)
  const int totalpixelbytes = width * height * bpp / 8;

  //temporary variables used inside the loops
  int i = 0; //loop variable for height
  int j = 0; //loop variable for width
  Uint32 * oldpixel; //pointer to old pixel
  Uint32 * newpixel; //pointer to new pixel
  int x = 0; //destination x-coordinate
  int y = 0; //destination y-coordinate

  Uint32 * pixelbuffer = (Uint32 *) malloc(totalpixelbytes); //holds pixels

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  memcpy((char *) pixelbuffer, (char *) image->pixels, totalpixelbytes);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      y = width - 1 - j;
      x = i;
      oldpixel = (Uint32 *) image->pixels + ((i * width) + j);
      newpixel = (Uint32 *) pixelbuffer + ((y * height) + x);
      *newpixel = *oldpixel;
    }
  }

  image = SDL_SetVideoMode(height, width, bpp, SDL_SWSURFACE);

  memcpy((char *) image->pixels, (char *) pixelbuffer, totalpixelbytes);

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */

  free(pixelbuffer);
}

void rotateRight(SDL_Surface * image, BoundedStack * history)
{
  const int width = image->w; //image width (original)
  const int height = image->h; //image height (original)
  const int bpp = image->format->BitsPerPixel; //bpp (normally 32)
  const int totalpixelbytes = width * height * bpp / 8;

  //temporary variables used inside the loops
  int i = 0; //loop variable for height
  int j = 0; //loop variable for width
  Uint32 * oldpixel; //pointer to old pixel
  Uint32 * newpixel; //pointer to new pixel
  int x = 0; //destination x-coordinate
  int y = 0; //destination y-coordinate

  Uint32 * pixelbuffer = (Uint32 *) malloc(totalpixelbytes); //holds pixels

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  memcpy((char *) pixelbuffer, (char *) image->pixels, totalpixelbytes);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      y = j;
      x = height - 1 - i;
      oldpixel = (Uint32 *) image->pixels + ((i * width) + j);
      newpixel = (Uint32 *) pixelbuffer + ((y * height) + x);
      *newpixel = *oldpixel;
    }
  }

  image = SDL_SetVideoMode(height, width, bpp, SDL_SWSURFACE);

  memcpy((char *) image->pixels, (char *) pixelbuffer, totalpixelbytes);

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */

  free(pixelbuffer);
} 

void flipHorizontal(SDL_Surface * image, BoundedStack * history)
{
  const int width = image->w; //image width (original)
  const int height = image->h; //image height (original)
  const int bpp = image->format->BitsPerPixel; //bpp (normally 32)
  const int totalpixelbytes = width * height * bpp / 8;

  //temporary variables used inside the loops
  int i = 0; //loop variable for height
  int j = 0; //loop variable for width
  Uint32 * oldpixel; //pointer to old pixel
  Uint32 * newpixel; //pointer to new pixel
  int x = 0; //destination x-coordinate

  Uint32 * pixelbuffer = (Uint32 *) malloc(totalpixelbytes); //holds pixels

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  memcpy((char *) pixelbuffer, (char *) image->pixels, totalpixelbytes);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      x = width - j - 1;
      oldpixel = (Uint32 *) image->pixels + ((i * width) + j);
      newpixel = (Uint32 *) pixelbuffer + ((i * width) + x);
      *newpixel = *oldpixel;
    }
  }

  memcpy((char *) image->pixels, (char *) pixelbuffer, totalpixelbytes);

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */

  free(pixelbuffer);
}

void flipVertical(SDL_Surface * image, BoundedStack * history)
{
  const int width = image->w; //image width (original)
  const int height = image->h; //image height (original)
  const int bpp = image->format->BitsPerPixel; //bpp (normally 32)
  const int totalpixelbytes = width * height * bpp / 8;

  //temporary variables used inside the loops
  int i = 0; //loop variable for height
  int j = 0; //loop variable for width
  Uint32 * oldpixel; //pointer to old pixel
  Uint32 * newpixel; //pointer to new pixel
  int y = 0; //destination y-coordinate

  Uint32 * pixelbuffer = (Uint32 *) malloc(width * height * bpp / 8); //holds pixels

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  memcpy((char *) pixelbuffer, (char *) image->pixels, totalpixelbytes);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      y = height - i - 1;
      oldpixel = (Uint32 *) image->pixels + ((i * width) + j);
      newpixel = (Uint32 *) pixelbuffer + ((y * width) + j);
      *newpixel = *oldpixel;
    }
  }

  memcpy((char *) image->pixels, (char *) pixelbuffer, width * height * bpp / 8);

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */

  free(pixelbuffer);
}

void negateImage(SDL_Surface * image, BoundedStack * history)
{
  /* image dimensions */
  const int width = image->w; /* image width */
  const int height = image->h; /* image height */

  /* temporary variables used inside the loops */
  int i = 0; /* loop variable for height */
  int j = 0; /* loop variable for width */
  Uint32 * pixel; /* pointer to pixel being manipulated */
  int r = 0; /* red component of pixel being manipulated */
  int g = 0; /* green component of pixel being manipulated */
  int b = 0; /* blue component of pixel being manipulated */
  int rmask = 0x000000FF; /* mask for red component */
  int gmask = 0x0000FF00; /* mask for green component */
  int bmask = 0x00FF0000; /* mask for blue component */

  /* masks used depend on the byte ordering of the system */
  if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  { /* initialised values assume little endian */
    rmask = 0xFF000000;
    gmask = 0x00FF0000;
    bmask = 0x0000FF00;
  }

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      pixel = (Uint32 *) image->pixels + ((i * width) + j);
      r = rmask - (*pixel & rmask);
      g = gmask - (*pixel & gmask);
      b = bmask - (*pixel & bmask);
      *pixel = r | g | b;
    }
  }

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */
}

void grayscaleImage(SDL_Surface * image, BoundedStack * history)
{
  /* image dimensions */
  const int width = image->w; /* image width */
  const int height = image->h; /* image height */
  const int xpp = image->format->BytesPerPixel; //bpp (normally 32)
  const int totalpixels = width * xpp * height;

  /* processing variables */
  int i = 0; /* loop variable for pixel index */
  Uint8 * pixel = NULL; /* pointer to pixel being manipulated */
  Uint8 * r = NULL; /* red component of pixel being manipulated */
  Uint8 * g = NULL; /* green component of pixel being manipulated */
  Uint8 * b = NULL; /* blue component of pixel being manipulated */
  Uint8 average; /* average of r,g,b used as grayscale pixel colour */
  int rpos = 2; /* position of red byte w.r.t. start of 4-byte chunk */
  int gpos = 1; /* position of green byte w.r.t. start of 4-byte chunk */
  int bpos = 0; /* position of blue byte w.r.t. start of 4-byte chunk */

  if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  { /* initialised values assume little endian */
    rpos = 0;
    gpos = 1;
    bpos = 2;
  }

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  startTimer();

  for (i = 0; i < totalpixels; i += xpp)
  {
      pixel = (Uint8 *) image->pixels + i;

      r = pixel + rpos;
      g = pixel + gpos;
      b = pixel + bpos;

      average = (*r + *g + *b) / 3;

      *r = average;
      *g = average;
      *b = average;
  }

  stopTimer();
  showElapsedTime();

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */
}

/* temporary... need to measure efficiency */

void grayscaleImageOld(SDL_Surface * image, BoundedStack * history)
{
  /* image dimensions */
  const int width = image->w; /* image width */
  const int height = image->h; /* image height */

  /* temporary variables used inside the loops */
  int i = 0; /* loop variable for height */
  int j = 0; /* loop variable for width */
  Uint32 * pixel; /* pointer to pixel being manipulated */
  int r = 0; /* red component of pixel being manipulated */
  int g = 0; /* green component of pixel being manipulated */
  int b = 0; /* blue component of pixel being manipulated */
  int rmask; /* mask for red component */
  int gmask; /* mask for green component */
  int bmask; /* mask for blue component */
  int average; /* average of r,g,b used as grayscale pixel colour */

  /* masks used depend on the byte ordering of the system */
  if (SDL_BYTEORDER == SDL_LIL_ENDIAN)
  {
    rmask = 0x000000FF;
    gmask = 0x0000FF00;
    bmask = 0x00FF0000;
  }
  else if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  {
    rmask = 0xFF000000;
    gmask = 0x00FF0000;
    bmask = 0x0000FF00;
  }

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  startTimer();

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      pixel = (Uint32 *) image->pixels + ((i * width) + j);

      if (SDL_BYTEORDER == SDL_LIL_ENDIAN)
      {
        r = (*pixel & rmask) >> 0;
        g = (*pixel & gmask) >> 8;
        b = (*pixel & bmask) >> 16;

        average = (r + g + b) / 3;

        r = (average << 0) & rmask;
        g = (average << 8) & gmask;
        b = (average << 16) & bmask;
      }
      else if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
      {
        r = (*pixel & rmask) >> 24;
        g = (*pixel & gmask) >> 16;
        b = (*pixel & bmask) >> 8;

        average = (r + g + b) / 3;

        r = (average << 24) & rmask;
        g = (average << 16) & gmask;
        b = (average << 8) & bmask;
      }

      *pixel = r | g | b;
    }
  }

  stopTimer();
  showElapsedTime();

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */
}

void sandstormImage(SDL_Surface * image, BoundedStack * history)
{
  /* image dimensions */
  const int width = image->w; /* image width */
  const int height = image->h; /* image height */

  /* temporary variables used inside the loops */
  int i = 0; /* loop variable for height */
  int j = 0; /* loop variable for width */
  Uint32 * pixel; /* pointer to pixel being manipulated */
  int r = 0; /* red component of pixel being manipulated */
  int g = 0; /* green component of pixel being manipulated */
  int b = 0; /* blue component of pixel being manipulated */
  int rmask; /* mask for red component */
  int gmask; /* mask for green component */
  int bmask; /* mask for blue component */
  int average; /* average of r,g,b used as grayscale pixel colour */

  /* masks used depend on the byte ordering of the system */
  if (SDL_BYTEORDER == SDL_LIL_ENDIAN)
  {
    rmask = 0x000000FF;
    gmask = 0x0000FF00;
    bmask = 0x00FF0000;
  }
  else if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  {
    rmask = 0xFF000000;
    gmask = 0x00FF0000;
    bmask = 0x0000FF00;
  }

  if (SDL_MUSTLOCK(image))
    SDL_LockSurface(image);

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      pixel = (Uint32 *) image->pixels + ((i * width) + j);

      if (SDL_BYTEORDER == SDL_LIL_ENDIAN)
      {
        r = *pixel / rmask >> 8;
        g = *pixel / gmask >> 8;
        b = *pixel / bmask >> 8;
      }
      else if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
      {
        r = *pixel / rmask << 8;
        g = *pixel / gmask << 8;
        b = *pixel / bmask << 8;
      }

      average = (r + g + b) / 3;
      r = average * rmask;
      g = average * gmask;
      b = average * bmask;

      *pixel = r | g | b;
    }
  }

  if (SDL_MUSTLOCK(image))
    SDL_UnlockSurface(image);

  SDL_Flip(image);
  historySave(image, history); /* save current state of image */
}
