#if defined(__WIN32)
  #include <winsock2.h>
#else
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <sys/types.h>
  #include <arpa/inet.h>
  #include <netdb.h>
#endif

#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "sock.h"

/* Connection interface */

int sockOpen(int protocol)
{
  #ifdef __WIN32__
    WSADATA wsaData; /* WSADATA structure to be filled by WSAStartup() */
    WORD winsockVersion = MAKEWORD(1, 1); /* use Winsock 1.1 */

    WSAStartup(winsockVersion, &wsaData);
  #endif
  return socket(AF_INET, protocol, 0);
}

int sockOpenTCP()
{
  return sockOpen(SOCK_STREAM);
}

int sockOpenUDP()
{
  return sockOpen(SOCK_DGRAM);
}

void sockClose(int sockfd)
{
  #ifndef __WIN32__
    shutdown(sockfd, SHUT_WR);
  #else
    WSACleanup();
  #endif
  close(sockfd);
}

int sockConnect(int sockfd, const char * target, unsigned short int port)
{
  unsigned int ip = (unsigned int) inet_addr(target);
  struct hostent * targetdns = NULL;
  struct sockaddr_in targetaddr;

  /* determine IP as given or from DNS name */
  if (ip == (unsigned int) (-1))
  { /* attempt DNS lookup */
    targetdns = gethostbyname(target);

    if (targetdns == NULL) /* ok it's invalid */
      return -1;
    else
      memcpy((unsigned long *) &ip, targetdns->h_addr, targetdns->h_length);

    if (ip == (unsigned int) (-1)) /* DNS lookup failed */
      return -1;
  }

  /* clean out address structure */
  memset((void *) &targetaddr, 0, sizeof(targetaddr));

  /* fill in address structure */
  targetaddr.sin_family = AF_INET;
  targetaddr.sin_addr.s_addr = ip;
  targetaddr.sin_port = htons(port);

  return connect(sockfd, (struct sockaddr *) &targetaddr, sizeof(targetaddr));
}

/* Data transfer interface */

int sockWrite(int sock, void * buf, size_t bytes)
{
  return send(sock, (char *) buf, bytes, 0);
}

int sockRead(int sock, void * buf, size_t bytes)
{
  return recv(sock, (char *) buf, bytes, 0);
}

int sockWriteAll(int sock, void * buf, size_t nbytes)
{ /* adapted from http://www.cas.mcmaster.ca/~qiao/courses/cs3mh3/
   *     tutorials/socket.html
   */
  int byteswritten = 0;
  int bytesleft = nbytes;
  char * pbuf = (char *) buf; /* to avoid void pointer arithmetic */
  while (bytesleft > 0)
  {
    if ((byteswritten = send(sock, (char *) pbuf, bytesleft, 0)) < 0) /* err */
      return byteswritten;
    bytesleft -= byteswritten; /* some bytes were written; update bytesleft */
    pbuf += byteswritten; /* update buf pointer */
  }
  return byteswritten;
}

int sockReadAll(int sock, void * buf, size_t nbytes)
{ /* adapted from http://www.cas.mcmaster.ca/~qiao/courses/cs3mh3/
   *     tutorials/socket.html
   */
  int bytesread = 0;
  int bytesleft = nbytes;
  char * pbuf = (char *) buf; /* to avoid void pointer arithmetic */
  while (bytesleft > 0)
  {
    if ((bytesread = recv(sock, (char *) pbuf, bytesleft, 0)) < 0) /* error */
      return bytesread;
    bytesleft -= bytesread; /* some bytes were read; update bytesleft */
    pbuf += bytesread; /* update buf pointer */
  }
  return bytesread;
}

int sockWriteStr(int sock, char * buf)
{
  return sockWriteAll(sock, buf, strlen(buf));
}

int sockWriteStrWithLen(int sock, char * buf)
{ /* msgsize will be stored in first byte; next 3 are redundant */
  int writereturn = 0; /* return value of writeToSocket() */
  int msgsize = strlen(buf) + 1; /* allow one for \0 */
  int nmsgsize = htonl(msgsize); /* msgsize in network byte order */
  char * temp = (char *) malloc(msgsize + 4); /* allow 4 bytes for msgsize */
  memcpy(&temp, &nmsgsize, 4);
  strcpy(temp + 4, buf); /* copy data into temp (skip first 4 bytes) */
  temp[msgsize + 3] = '\0'; /* null-terminate temp */
  writereturn = sockWriteAll(sock, temp, msgsize + 4); /* send temp */
  free(temp);
  return writereturn;
}

int sockReadStrWithLen(int sock, char * buf)
{
  int msgsize = 0;
  /* On Athena there is a strange thing happening... if we try to
   * read just one byte it gets stuck. So we add an extra 3 bytes
   * for fun so that Athena is happy. This little 'feature' took
   * a whole day to spot.
   */
  sockRead(sock, &msgsize, 4); /* get msgsize stored in first byte */
  msgsize = ntohl(msgsize);
  return sockReadAll(sock, buf, msgsize);
}
